@@ -34,7 +34,7 @@
 void
 error_print(int type, int line, const char *error)
 {
-	char *w = NULL;
+	const char *w = NULL;
 
 	switch(type) {
 	case GOB_WARN:
@@ -77,6 +77,11 @@ remove_sep(const char *base)
 {
 	char *p;
 	char *s = g_strdup(base);
+
+	/* don't eat C++ :: thingies */
+	if (for_cpp && strstr (s, "::") != NULL)
+		return s;
+
 	while((p = strchr(s, ':')))
 		strcpy(p, p+1);
 	return s;
@@ -87,6 +92,11 @@ replace_sep(const char *base, char r)
 {
 	char *p;
 	char *s = g_strdup(base);
+
+	/* don't eat C++ :: thingies */
+	if (for_cpp && strstr (s, "::") != NULL)
+		return s;
+
 	while((p=strchr(s,':')))
 		*p = r;
 	if(*s == r) {
@@ -104,7 +114,15 @@ separns_replace_sep(const char *base, char **ns, char **name, char r)
 {
 	char *p;
 	char *s = g_strdup(base);
+
 	*ns = NULL;
+	
+	/* don't eat C++ :: thingies */
+	if (for_cpp && strstr (s, "::") != NULL) {
+		*name = s;
+		return;
+	}
+
 	if((p=strchr(s,':')) && p!=s) {
 		*p = '\0';
 		*ns = g_strdup(s);
@@ -164,26 +182,39 @@ typedef struct _OurGtkType OurGtkType;
 struct _OurGtkType {
 	gboolean simple;
 	char *gtkname;
-	char *typename;
+	char *cast;
+	char *type_name;
+	char *type_pointer;
 	int special;
 };
 const OurGtkType our_gtk_type_table[] = {
-	{ TRUE, "NONE",		"void ", -1 },
-	{ TRUE, "CHAR",		"gchar ", -1 },
-	{ TRUE, "UCHAR",	"guchar ", -1 },
-	{ TRUE, "BOOL",		"gboolean ", -1 },
-	{ TRUE, "INT",		"gint ", -1 },
-	{ TRUE, "UINT",		"guint ", -1 },
-	{ TRUE, "LONG",		"glong ", -1 },
-	{ TRUE, "ULONG",	"gulong ", -1 },
-	{ TRUE, "FLOAT",	"gfloat ", -1 },
-	{ TRUE, "DOUBLE",	"gdouble ", -1 },
-	{ TRUE, "STRING",	/*"GtkString"*/"gchar *", -1 },
-	{ TRUE, "ENUM",		/*"GtkEnum"*/"gint ", -1 },
-	{ TRUE, "FLAGS",	/*"GtkFlags"*/"guint ", -1 },
-	{ TRUE, "BOXED",	/*"GtkBoxed"*/"gpointer ", -1 },
-	{ TRUE, "POINTER",	"gpointer ", -1 },
-	{ TRUE, "OBJECT",	"GtkObject *", -1 },
+	{ TRUE, "NONE",		"void ",	"void",		NULL,	-1 },
+	{ TRUE, "CHAR",		"gchar ",	"gchar",	NULL,	-1 },
+	{ TRUE, "UCHAR",	"guchar ",	"guchar",	NULL,	-1 },
+	{ TRUE, "UNICHAR",	"gunichar ",	"gunichar",	NULL,	-1 },
+	{ TRUE, "BOOLEAN",	"gboolean ",	"gboolean",	NULL,	-1 },
+	{ TRUE, "INT",		"gint ",	"gint",		NULL,	-1 },
+	{ TRUE, "UINT",		"guint ", 	"guint",	NULL,	-1 },
+	{ TRUE, "LONG",		"glong ",	"glong",	NULL,	-1 },
+	{ TRUE, "ULONG",	"gulong ",	"gulong",	NULL,	-1 },
+	{ TRUE, "INT64",	"gint64 ",	"gint64",	NULL,	-1 },
+	{ TRUE, "UINT64",	"guint64 ", 	"guint64",	NULL,	-1 },
+	{ TRUE, "ENUM",		/*"enum"*/"gint ", "gint",	NULL,	-1 },
+	{ TRUE, "FLAGS",	/*"flags"*/"guint ", "guint",	NULL,	-1 },
+	{ TRUE, "FLOAT",	"gfloat ",	"gfloat",	NULL,	-1 },
+	{ TRUE, "DOUBLE",	"gdouble ",	"gdouble",	NULL,	-1 },
+	{ TRUE, "STRING",	/*"string"*/"gchar *", "gchar",	"*",	-1 },
+	{ TRUE, "POINTER",	"gpointer ",	"gpointer",	NULL,	-1 },
+	{ TRUE, "BOXED",	/*"boxed"*/"gpointer ", "gpointer", NULL, -1 },
+	{ TRUE, "OBJECT",	"GObject *",	"GObject",	"*",	-1 },
+	{ TRUE, "PARAM",	"GParamSpec *",	"GParamSpec",	"*",	-1 },
+
+	/* FIXME: VALUE_ARRAY, CLOSURE */
+	/* Note that those have some issues with g_value_ calls etc... so
+	 * we can't just add them */
+
+	/* Do we need this??? */
+#if 0
 	{ FALSE, "SIGNAL",	/*"GtkSignal"*/"___twopointertype ",
 		SPECIAL_2POINTER },
 	{ FALSE, "ARGS",	/*"GtkArgs"*/"___intpointertype ",
@@ -194,6 +225,7 @@ const OurGtkType our_gtk_type_table[] = {
 		SPECIAL_2POINTER },
 	{ FALSE, "FOREIGN",	/*"GtkForeign"*/"___twopointertype ",
 		SPECIAL_2POINTER },
+#endif
 
 	{ FALSE, NULL, NULL }
 };
@@ -216,24 +248,47 @@ init_type_hash(void)
 }
 
 const char *
-get_cast(char *type, gboolean simple_only)
+get_cast (const char *type, gboolean simple_only)
 {
 	OurGtkType *gtype;
 
-	init_type_hash();
+	init_type_hash ();
 
-	gtype = g_hash_table_lookup(type_hash, type);
+	gtype = g_hash_table_lookup (type_hash, type);
 
-	if(!gtype ||
-	   (simple_only &&
-	    !gtype->simple))
+	if (gtype == NULL ||
+	    (simple_only &&
+	     ! gtype->simple))
 		return NULL;
 
-	return gtype->typename;
+	return gtype->cast;
+}
+
+Type *
+get_tree_type (const char *type, gboolean simple_only)
+{
+	OurGtkType *gtype;
+	Node *node;
+
+	init_type_hash ();
+
+	gtype = g_hash_table_lookup (type_hash, type);
+
+	if (gtype == NULL ||
+	    (simple_only &&
+	     ! gtype->simple))
+		return NULL;
+
+	node = node_new (TYPE_NODE,
+			 "name", gtype->type_name,
+			 "pointer", gtype->type_pointer,
+			 NULL);
+
+	return (Type *)node;
 }
 
 static void
-mask_special_array(char *type, gboolean *special_array, gboolean *any_special)
+mask_special_array (const char *type, gboolean *special_array, gboolean *any_special)
 {
 	OurGtkType *gtype;
 
@@ -277,14 +332,37 @@ setup_special_array(Class *c, gboolean *special_array)
 	return any_special;
 }
 
-/* get the id without the first underscore, but only if we're removing them */
-const char *
-get_real_id(const char *id)
+char *
+get_type (const Type *t, gboolean postfix_to_stars)
 {
-	if( ! no_kill_underscores &&
-	    id[0] == '_' &&
-	    id[1] != '\0')
-		return &id[1];
-	else
-		return id;
+	char *s;
+	int i;
+	int extra;
+	GString *gs;
+
+	s = remove_sep(t->name);
+	gs = g_string_new(s);
+	g_free(s);
+
+	extra = 0;
+	if (postfix_to_stars) {
+		const char *p;
+		/*XXX: this is ugly perhaps we can do this whole postfix thing
+		  in a nicer way, we just count the number of '[' s and from
+		  that we deduce the number of dimensions, so that we can print
+		  that many stars */
+		for (p = t->postfix; p && *p; p++)
+			if(*p == '[') extra++;
+	}
+	g_string_append_c(gs, ' ');
+
+	if (t->pointer != NULL) {
+		g_string_append (gs, t->pointer);
+		for (i=0; i < extra; i++)
+			g_string_append_c (gs, '*');
+		g_string_append_c (gs, ' ');
+	}
+	
+	return g_string_free (gs, FALSE);
 }
+
